//-----------------------------------------------------------------------------
// MK3_2560_FASTFAT16
// Version 1.0
// 05.07.2012 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board
// mit ATMega-2560 @ 16MHz
// SD-Karte Ein-/Ausgabe (FAT16-Fileformat)
// (dieser Treiber ist schneller als der "MK3_2560_FAT16.h")
// Lesegeschwindigkeit ca. 125 kByte/sec
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_FASTFAT16.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : Port-B (SPI = Bit0 bis Bit3)
//-----------------------------------------------------------------------------
//
// Hinweis :
// 
// > Der Zugriff auf die SD-Karte erfolgt ueber eine Struktur :
//
//  Name = "MK3_SD"
//
//  MK3_SD.partition
//  MK3_SD.fat
//  MK3_SD.root
//  MK3_SD.dir
//  MK3_SD.file
//
//  > die Struktur wird am Ende dieses Treibers initialisiert
//    und kann bei bedarf erweitert werden
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Copyright Hinweis von UB : 
// ich habe den Code fuer das FAT16-Dateisystem von dieser Quelle :
// Adr : http://www.roland-riegel.de/sd-reader/
// Name : MMC/SD/SDHC card library
// Ver : vom 12.06.2012
// Autor : Roland-Riegel
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Copyright (c) 2006-2012 by Roland Riegel <feedback@roland-riegel.de>
//
// This file is free software; you can redistribute it and/or modify
// it under the terms of either the GNU General Public License version 2
// or the GNU Lesser General Public License version 2.1, both as
// published by the Free Software Foundation.
//
//-----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Anpassungen von UB
// SD_RAW.c           : "sd_raw_available()" -> return wert immer 1
//                      "sd_raw_locked()"    -> return wert immer 0
//                      "sd_raw_init()" -> "configure_pin_available" entfernt
//                      "sd_raw_init()" -> "configure_pin_locked" entfernt
//                      SDHC Unterstuetzung entfernt
// SD_RAW.h           : keine
// SD_RAW_CONFIG.h    : nur ATMega2560 support (Port B0-B3)
//                      "configure_pin_available" entfernt
//                      "configure_pin_locked" entfernt
//                      "get_pin_available" entfernt
//                      "get_pin_locked" entfernt
//                      SDHC Unterstuetzung entfernt
// PARTITION.c        : DYNAMIC_MEMORY Unterstuetzung entfernt
// PARTITION.h        : keine
// PARTITION_CONFIG.h : keine
// BYTEORDERING.c     : SWAP_NEEDED Unterstuetzung entfernt
// BYTEORDERING.h     : SWAP_NEEDED Unterstuetzung entfernt
// SD_READER_CONFIG.h : DYNAMIC_MEMORY Unterstuetzung entfernt
// FAT.c              : FAT32 Unterstuetzung entfernt
//                      DYNAMIC_MEMORY Unterstuetzung entfernt
//                      DATETIME Unterstuetzung entfernt
//                      "fat_dir_entry_read_callback()" -> cast auf pointer
// FAT.h              : UINTPTR_MAX definiert -> UINTPTR_MAX_UB
//                      DATETIME Unterstuetzung entfernt
// FAT_CONFIG.h       : FAT32 Unterstuetzung entfernt
//                      DATETIME Unterstuetzung entfernt
//----------------------------------------------------------------------------
//
// Hinweise :
// 
//   > Speicherplatz -> mit allem = ca. 18 kByte
//                      ohne Write = ca. 11 kByte 
//                      minimal = ca. 9 kByte 
//
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------

#include <string.h>
#include <avr/io.h>

//-----------------------------------------------------------------------------
// Hier koennen Funktionen die nicht benutzt werden
// abgeschaltet werden...einfach den DEFINE von "1" auf "0" aendern
//-----------------------------------------------------------------------------

#define USE_MK3_SD_WRITE 1
#define USE_MK3_SD_GET_INFO 1
#define USE_MK3_SD_SEEK_FILE 1
#define USE_MK3_SD_GET_FS_SIZE 1
#define USE_MK3_SD_GET_FS_FREE 1


//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------

#define MK3_DDR_FASTSD DDRB      // DDR-Register der SD-Karte
#define MK3_PORT_FASTSD PORTB	 // Port der SD-Karte

#define MK3_FASTSD_SlaveSel 0x01 // SlaveSelect der SD-Karte (PB0)
#define MK3_FASTSD_SCK      0x02 // SCK der SD-Karte         (PB1)
#define MK3_FASTSD_MOSI     0x04 // MOSI der SD-Karte        (PB2)
#define MK3_FASTSD_MISO     0x08 // MISO der SD-Karte        (PB3)



//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
//
// MK3_SD_READ_FILE_IN_ROOT();   // oeffnet ein File zum lesen
// MK3_SD_WRITE_FILE_IN_ROOT();  // oeffnet ein File zum schreiben
// MK3_SD_CLOSE_FILE();          // schliesst ein File
//
// sd_raw_init();                // init und Check ob eine SD-Karte vorhanden ist
// partition_open();             // Partition oeffnen
// fat_open();                   // FAT oeffnen
// fat_get_dir_entry_of_path();  // ROOT oeffnen
// fat_open_dir();               // Directory oeffnen
// open_file_in_dir();           // File oeffnen
// fat_read_file();              // aus dem File lesen (in einen Puffer)
// fat_write_file();             // in das File schreiben (von einem Puffer)
// fat_close_file();             // File schliessen
// fat_close_dir();              // Directory schliessen
// fat_close();                  // FAT schliessen
// partition_close();            // Partition schliessen
//
// find_file_in_dir();           // Test ob File vorhanden ist
// fat_read_dir();               // Directory auslesen
// fat_delete_file();            // File loeschen
// fat_create_file();            // File anlegen
// fat_create_dir();             // Direktory anlegen
// fat_move_file();              // File verschieben
// fat_seek_file();              // Schreib Pointer verschieben
//
//-----------------------------------------------------------------------------




//----------------------------------------------------------------------------
// SD_RAW_CONFIG.h [START]
//----------------------------------------------------------------------------

#define SD_RAW_WRITE_SUPPORT USE_MK3_SD_WRITE
#define SD_RAW_WRITE_BUFFERING 1
#define SD_RAW_SAVE_RAM 0

#define configure_pin_mosi() MK3_DDR_FASTSD |= MK3_FASTSD_MOSI
#define configure_pin_sck() MK3_DDR_FASTSD |= MK3_FASTSD_SCK
#define configure_pin_ss() MK3_DDR_FASTSD |= MK3_FASTSD_SlaveSel
#define configure_pin_miso() MK3_DDR_FASTSD &= ~MK3_FASTSD_MISO

#define select_card() MK3_PORT_FASTSD &= ~MK3_FASTSD_SlaveSel
#define unselect_card() MK3_PORT_FASTSD |= MK3_FASTSD_SlaveSel

typedef uint32_t offset_t;


#if SD_RAW_WRITE_SUPPORT
  #undef SD_RAW_SAVE_RAM
  #define SD_RAW_SAVE_RAM 0
#else
  #undef SD_RAW_WRITE_BUFFERING
  #define SD_RAW_WRITE_BUFFERING 0
#endif

//----------------------------------------------------------------------------
// SD_RAW_CONFIG.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// SD_RAW.h [START]
//----------------------------------------------------------------------------

#define SD_RAW_FORMAT_HARDDISK 0
#define SD_RAW_FORMAT_SUPERFLOPPY 1
#define SD_RAW_FORMAT_UNIVERSAL 2
#define SD_RAW_FORMAT_UNKNOWN 3


struct sd_raw_info
{
    uint8_t manufacturer;
    uint8_t oem[3];
    uint8_t product[6];
    uint8_t revision;
    uint32_t serial;
    uint8_t manufacturing_year;
    uint8_t manufacturing_month;
    offset_t capacity;
    uint8_t flag_copy;
    uint8_t flag_write_protect;
    uint8_t flag_write_protect_temp;
    uint8_t format;
};

typedef uint8_t (*sd_raw_read_interval_handler_t)(uint8_t* buffer, offset_t offset, void* p);
typedef uintptr_t (*sd_raw_write_interval_handler_t)(uint8_t* buffer, offset_t offset, void* p);

uint8_t sd_raw_init();
uint8_t sd_raw_available();
uint8_t sd_raw_locked();

uint8_t sd_raw_read(offset_t offset, uint8_t* buffer, uintptr_t length);
uint8_t sd_raw_read_interval(offset_t offset, uint8_t* buffer, uintptr_t interval, uintptr_t length, sd_raw_read_interval_handler_t callback, void* p);
uint8_t sd_raw_write(offset_t offset, const uint8_t* buffer, uintptr_t length);
uint8_t sd_raw_write_interval(offset_t offset, uint8_t* buffer, uintptr_t length, sd_raw_write_interval_handler_t callback, void* p);
uint8_t sd_raw_sync();

uint8_t sd_raw_get_info(struct sd_raw_info* info);

//----------------------------------------------------------------------------
// SD_RAW.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// PARTITION_CONFIG.h [START]
//----------------------------------------------------------------------------

#define PARTITION_COUNT 1

//----------------------------------------------------------------------------
// PARTITION_CONFIG.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// PARTITION.h [START]
//----------------------------------------------------------------------------

#define PARTITION_TYPE_FREE 0x00
#define PARTITION_TYPE_FAT12 0x01
#define PARTITION_TYPE_FAT16_32MB 0x04
#define PARTITION_TYPE_EXTENDED 0x05
#define PARTITION_TYPE_FAT16 0x06
#define PARTITION_TYPE_FAT32 0x0b
#define PARTITION_TYPE_FAT32_LBA 0x0c
#define PARTITION_TYPE_FAT16_LBA 0x0e
#define PARTITION_TYPE_EXTENDED_LBA 0x0f
#define PARTITION_TYPE_UNKNOWN 0xff

typedef uint8_t (*device_read_t)(offset_t offset, uint8_t* buffer, uintptr_t length);
typedef uint8_t (*device_read_callback_t)(uint8_t* buffer, offset_t offset, void* p);
typedef uint8_t (*device_read_interval_t)(offset_t offset, uint8_t* buffer, uintptr_t interval, uintptr_t length, device_read_callback_t callback, void* p);
typedef uint8_t (*device_write_t)(offset_t offset, const uint8_t* buffer, uintptr_t length);
typedef uintptr_t (*device_write_callback_t)(uint8_t* buffer, offset_t offset, void* p);
typedef uint8_t (*device_write_interval_t)(offset_t offset, uint8_t* buffer, uintptr_t length, device_write_callback_t callback, void* p);

struct partition_struct
{
    device_read_t device_read;
    device_read_interval_t device_read_interval;
    device_write_t device_write;
    device_write_interval_t device_write_interval;
    uint8_t type;
    uint32_t offset;
    uint32_t length;
};

struct partition_struct* partition_open(device_read_t device_read, device_read_interval_t device_read_interval, device_write_t device_write, device_write_interval_t device_write_interval, int8_t index);
uint8_t partition_close(struct partition_struct* partition);

//----------------------------------------------------------------------------
// PARTITION.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// BYTEORDERING.h [START]
//----------------------------------------------------------------------------

#define HTOL16(val) (val)
#define HTOL32(val) (val)
#define LTOH16(val) (val)
#define LTOH32(val) (val)

#define htol16(h) (h)
#define htol32(h) (h)
#define ltoh16(l) (l)
#define ltoh32(l) (l)

uint16_t read16(const uint8_t* p);
uint32_t read32(const uint8_t* p);
void write16(uint8_t* p, uint16_t i);
void write32(uint8_t* p, uint32_t i);

//----------------------------------------------------------------------------
// BYTEORDERING.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// SD_READER_CONFIG.h [START]
//----------------------------------------------------------------------------

// leer

//----------------------------------------------------------------------------
// SD_READER_CONFIG.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// FAT_CONFIG.h [START]
//----------------------------------------------------------------------------

#define FAT_WRITE_SUPPORT SD_RAW_WRITE_SUPPORT
#define FAT_LFN_SUPPORT 1
#define FAT_DELAY_DIRENTRY_UPDATE 0

#define FAT_FS_COUNT 1
#define FAT_FILE_COUNT 1
#define FAT_DIR_COUNT 2

typedef uint16_t cluster_t;

//----------------------------------------------------------------------------
// FAT_CONFIG.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// FAT.h [START]
//----------------------------------------------------------------------------

#define UINTPTR_MAX_UB 0xFFFF

#define FAT_ATTRIB_READONLY (1 << 0)
#define FAT_ATTRIB_HIDDEN (1 << 1)
#define FAT_ATTRIB_SYSTEM (1 << 2)
#define FAT_ATTRIB_VOLUME (1 << 3)
#define FAT_ATTRIB_DIR (1 << 4)
#define FAT_ATTRIB_ARCHIVE (1 << 5)

#define FAT_SEEK_SET 0
#define FAT_SEEK_CUR 1
#define FAT_SEEK_END 2

struct partition_struct;
struct fat_fs_struct;
struct fat_file_struct;
struct fat_dir_struct;

struct fat_dir_entry_struct
{
    char long_name[32];
    uint8_t attributes;
    cluster_t cluster;   
    uint32_t file_size;   
    offset_t entry_offset;
};

struct fat_fs_struct* fat_open(struct partition_struct* partition);
void fat_close(struct fat_fs_struct* fs);

struct fat_file_struct* fat_open_file(struct fat_fs_struct* fs, const struct fat_dir_entry_struct* dir_entry);
void fat_close_file(struct fat_file_struct* fd);
intptr_t fat_read_file(struct fat_file_struct* fd, uint8_t* buffer, uintptr_t buffer_len);
intptr_t fat_write_file(struct fat_file_struct* fd, const uint8_t* buffer, uintptr_t buffer_len);
uint8_t fat_seek_file(struct fat_file_struct* fd, int32_t* offset, uint8_t whence);
uint8_t fat_resize_file(struct fat_file_struct* fd, uint32_t size);

struct fat_dir_struct* fat_open_dir(struct fat_fs_struct* fs, const struct fat_dir_entry_struct* dir_entry);
void fat_close_dir(struct fat_dir_struct* dd);
uint8_t fat_read_dir(struct fat_dir_struct* dd, struct fat_dir_entry_struct* dir_entry);
uint8_t fat_reset_dir(struct fat_dir_struct* dd);

uint8_t fat_create_file(struct fat_dir_struct* parent, const char* file, struct fat_dir_entry_struct* dir_entry);
uint8_t fat_delete_file(struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry);
uint8_t fat_move_file(struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry, struct fat_dir_struct* parent_new, const char* file_new);
uint8_t fat_create_dir(struct fat_dir_struct* parent, const char* dir, struct fat_dir_entry_struct* dir_entry);
#define fat_delete_dir fat_delete_file
#define fat_move_dir fat_move_file

uint8_t fat_get_dir_entry_of_path(struct fat_fs_struct* fs, const char* path, struct fat_dir_entry_struct* dir_entry);

offset_t fat_get_fs_size(const struct fat_fs_struct* fs);
offset_t fat_get_fs_free(const struct fat_fs_struct* fs);

//----------------------------------------------------------------------------
// FAT.h [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// SD_RAW.c [START]
//----------------------------------------------------------------------------

#define CMD_GO_IDLE_STATE 0x00
#define CMD_SEND_OP_COND 0x01
#define CMD_SEND_IF_COND 0x08
#define CMD_SEND_CSD 0x09
#define CMD_SEND_CID 0x0a
#define CMD_STOP_TRANSMISSION 0x0c
#define CMD_SEND_STATUS 0x0d
#define CMD_SET_BLOCKLEN 0x10
#define CMD_READ_SINGLE_BLOCK 0x11
#define CMD_READ_MULTIPLE_BLOCK 0x12
#define CMD_WRITE_SINGLE_BLOCK 0x18
#define CMD_WRITE_MULTIPLE_BLOCK 0x19
#define CMD_PROGRAM_CSD 0x1b
#define CMD_SET_WRITE_PROT 0x1c
#define CMD_CLR_WRITE_PROT 0x1d
#define CMD_SEND_WRITE_PROT 0x1e
#define CMD_TAG_SECTOR_START 0x20
#define CMD_TAG_SECTOR_END 0x21
#define CMD_UNTAG_SECTOR 0x22
#define CMD_TAG_ERASE_GROUP_START 0x23
#define CMD_TAG_ERASE_GROUP_END 0x24
#define CMD_UNTAG_ERASE_GROUP 0x25
#define CMD_ERASE 0x26
#define CMD_SD_SEND_OP_COND 0x29
#define CMD_LOCK_UNLOCK 0x2a
#define CMD_APP 0x37
#define CMD_READ_OCR 0x3a
#define CMD_CRC_ON_OFF 0x3b


#define R1_IDLE_STATE 0
#define R1_ERASE_RESET 1
#define R1_ILL_COMMAND 2
#define R1_COM_CRC_ERR 3
#define R1_ERASE_SEQ_ERR 4
#define R1_ADDR_ERR 5
#define R1_PARAM_ERR 6

#define R2_CARD_LOCKED 0
#define R2_WP_ERASE_SKIP 1
#define R2_ERR 2
#define R2_CARD_ERR 3
#define R2_CARD_ECC_FAIL 4
#define R2_WP_VIOLATION 5
#define R2_INVAL_ERASE 6
#define R2_OUT_OF_RANGE 7
#define R2_CSD_OVERWRITE 7
#define R2_IDLE_STATE (R1_IDLE_STATE + 8)
#define R2_ERASE_RESET (R1_ERASE_RESET + 8)
#define R2_ILL_COMMAND (R1_ILL_COMMAND + 8)
#define R2_COM_CRC_ERR (R1_COM_CRC_ERR + 8)
#define R2_ERASE_SEQ_ERR (R1_ERASE_SEQ_ERR + 8)
#define R2_ADDR_ERR (R1_ADDR_ERR + 8)
#define R2_PARAM_ERR (R1_PARAM_ERR + 8)

#define R3_OCR_MASK (0xffffffffUL)
#define R3_IDLE_STATE (R1_IDLE_STATE + 32)
#define R3_ERASE_RESET (R1_ERASE_RESET + 32)
#define R3_ILL_COMMAND (R1_ILL_COMMAND + 32)
#define R3_COM_CRC_ERR (R1_COM_CRC_ERR + 32)
#define R3_ERASE_SEQ_ERR (R1_ERASE_SEQ_ERR + 32)
#define R3_ADDR_ERR (R1_ADDR_ERR + 32)
#define R3_PARAM_ERR (R1_PARAM_ERR + 32)

#define DR_STATUS_MASK 0x0e
#define DR_STATUS_ACCEPTED 0x05
#define DR_STATUS_CRC_ERR 0x0a
#define DR_STATUS_WRITE_ERR 0x0c


#define SD_RAW_SPEC_1 0
#define SD_RAW_SPEC_2 1
#define SD_RAW_SPEC_SDHC 2

#if !SD_RAW_SAVE_RAM
  static uint8_t raw_block[512];
  static offset_t raw_block_address;
  #if SD_RAW_WRITE_BUFFERING
    static uint8_t raw_block_written;
  #endif
#endif


static uint8_t sd_raw_card_type;


static void sd_raw_send_byte(uint8_t b);
static uint8_t sd_raw_rec_byte();
static uint8_t sd_raw_send_command(uint8_t command, uint32_t arg);

//----------------------------------------------------------------------------
uint8_t sd_raw_init()
{
    /* enable outputs for MOSI, SCK, SS, input for MISO */
    configure_pin_mosi();
    configure_pin_sck();
    configure_pin_ss();
    configure_pin_miso();

    unselect_card();

    /* initialize SPI with lowest frequency; max. 400kHz during identification mode of card */
    SPCR = (0 << SPIE) | /* SPI Interrupt Enable */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (1 << SPR1) | /* Clock Frequency: f_OSC / 128 */
           (1 << SPR0);
    SPSR &= ~(1 << SPI2X); /* No doubled clock frequency */

    /* initialization procedure */
    sd_raw_card_type = 0;
    
    if(!sd_raw_available())
        return 0;

    /* card needs 74 cycles minimum to start up */
    for(uint8_t i = 0; i < 10; ++i)
    {
        /* wait 8 clock cycles */
        sd_raw_rec_byte();
    }

    /* address card */
    select_card();

    /* reset card */
    uint8_t response;
    for(uint16_t i = 0; ; ++i)
    {
        response = sd_raw_send_command(CMD_GO_IDLE_STATE, 0);
        if(response == (1 << R1_IDLE_STATE))
            break;

        if(i == 0x1ff)
        {
            unselect_card();
            return 0;
        }
    }

    {
        /* determine SD/MMC card type */
        sd_raw_send_command(CMD_APP, 0);
        response = sd_raw_send_command(CMD_SD_SEND_OP_COND, 0);
        if((response & (1 << R1_ILL_COMMAND)) == 0)
        {
            /* card conforms to SD 1 card specification */
            sd_raw_card_type |= (1 << SD_RAW_SPEC_1);
        }
        else
        {
            /* MMC card */
        }
    }

    /* wait for card to get ready */
    for(uint16_t i = 0; ; ++i)
    {
        if(sd_raw_card_type & ((1 << SD_RAW_SPEC_1) | (1 << SD_RAW_SPEC_2)))
        {
            uint32_t arg = 0;

            sd_raw_send_command(CMD_APP, 0);
            response = sd_raw_send_command(CMD_SD_SEND_OP_COND, arg);
        }
        else
        {
            response = sd_raw_send_command(CMD_SEND_OP_COND, 0);
        }

        if((response & (1 << R1_IDLE_STATE)) == 0)
            break;

        if(i == 0x7fff)
        {
            unselect_card();
            return 0;
        }
    }

    /* set block size to 512 bytes */
    if(sd_raw_send_command(CMD_SET_BLOCKLEN, 512))
    {
        unselect_card();
        return 0;
    }

    /* deaddress card */
    unselect_card();

    /* switch to highest SPI frequency possible */
    SPCR &= ~((1 << SPR1) | (1 << SPR0)); /* Clock Frequency: f_OSC / 4 */
    SPSR |= (1 << SPI2X); /* Doubled Clock Frequency: f_OSC / 2 */

#if !SD_RAW_SAVE_RAM
    /* the first block is likely to be accessed first, so precache it here */
    raw_block_address = (offset_t) -1;
#if SD_RAW_WRITE_BUFFERING
    raw_block_written = 1;
#endif
    if(!sd_raw_read(0, raw_block, sizeof(raw_block)))
        return 0;
#endif

    return 1;
}

//----------------------------------------------------------------------------
uint8_t sd_raw_available()
{
    return 0x01; // Karte immer vorhanden
}

//----------------------------------------------------------------------------
uint8_t sd_raw_locked()
{
    return 0x00; // Karte nie schreibgeschuetzt
}

//----------------------------------------------------------------------------
void sd_raw_send_byte(uint8_t b)
{
    SPDR = b;
    /* wait for byte to be shifted out */
    while(!(SPSR & (1 << SPIF)));
    SPSR &= ~(1 << SPIF);
}

//----------------------------------------------------------------------------
uint8_t sd_raw_rec_byte()
{
    /* send dummy data for receiving some */
    SPDR = 0xff;
    while(!(SPSR & (1 << SPIF)));
    SPSR &= ~(1 << SPIF);

    return SPDR;
}

//----------------------------------------------------------------------------
uint8_t sd_raw_send_command(uint8_t command, uint32_t arg)
{
    uint8_t response;

    /* wait some clock cycles */
    sd_raw_rec_byte();

    /* send command via SPI */
    sd_raw_send_byte(0x40 | command);
    sd_raw_send_byte((arg >> 24) & 0xff);
    sd_raw_send_byte((arg >> 16) & 0xff);
    sd_raw_send_byte((arg >> 8) & 0xff);
    sd_raw_send_byte((arg >> 0) & 0xff);
    switch(command)
    {
        case CMD_GO_IDLE_STATE:
           sd_raw_send_byte(0x95);
           break;
        case CMD_SEND_IF_COND:
           sd_raw_send_byte(0x87);
           break;
        default:
           sd_raw_send_byte(0xff);
           break;
    }
    
    /* receive response */
    for(uint8_t i = 0; i < 10; ++i)
    {
        response = sd_raw_rec_byte();
        if(response != 0xff)
            break;
    }

    return response;
}

//----------------------------------------------------------------------------
uint8_t sd_raw_read(offset_t offset, uint8_t* buffer, uintptr_t length)
{
    offset_t block_address;
    uint16_t block_offset;
    uint16_t read_length;
    while(length > 0)
    {
        /* determine byte count to read at once */
        block_offset = offset & 0x01ff;
        block_address = offset - block_offset;
        read_length = 512 - block_offset; /* read up to block border */
        if(read_length > length)
            read_length = length;
        
#if !SD_RAW_SAVE_RAM
        /* check if the requested data is cached */
        if(block_address != raw_block_address)
#endif
        {
#if SD_RAW_WRITE_BUFFERING
            if(!sd_raw_sync())
                return 0;
#endif

            /* address card */
            select_card();

            /* send single block request */

            if(sd_raw_send_command(CMD_READ_SINGLE_BLOCK, block_address))
            {
                unselect_card();
                return 0;
            }

            /* wait for data block (start byte 0xfe) */
            while(sd_raw_rec_byte() != 0xfe);

#if SD_RAW_SAVE_RAM
            /* read byte block */
            uint16_t read_to = block_offset + read_length;
            for(uint16_t i = 0; i < 512; ++i)
            {
                uint8_t b = sd_raw_rec_byte();
                if(i >= block_offset && i < read_to)
                    *buffer++ = b;
            }
#else
            /* read byte block */
            uint8_t* cache = raw_block;
            for(uint16_t i = 0; i < 512; ++i)
                *cache++ = sd_raw_rec_byte();
            raw_block_address = block_address;

            memcpy(buffer, raw_block + block_offset, read_length);
            buffer += read_length;
#endif
            
            /* read crc16 */
            sd_raw_rec_byte();
            sd_raw_rec_byte();
            
            /* deaddress card */
            unselect_card();

            /* let card some time to finish */
            sd_raw_rec_byte();
        }
#if !SD_RAW_SAVE_RAM
        else
        {
            /* use cached data */
            memcpy(buffer, raw_block + block_offset, read_length);
            buffer += read_length;
        }
#endif

        length -= read_length;
        offset += read_length;
    }

    return 1;
}

//----------------------------------------------------------------------------
uint8_t sd_raw_read_interval(offset_t offset, uint8_t* buffer, uintptr_t interval, uintptr_t length, sd_raw_read_interval_handler_t callback, void* p)
{
    if(!buffer || interval == 0 || length < interval || !callback)
        return 0;

#if !SD_RAW_SAVE_RAM
    while(length >= interval)
    {
        /* as reading is now buffered, we directly
         * hand over the request to sd_raw_read()
         */
        if(!sd_raw_read(offset, buffer, interval))
            return 0;
        if(!callback(buffer, offset, p))
            break;
        offset += interval;
        length -= interval;
    }

    return 1;
#else
    /* address card */
    select_card();

    uint16_t block_offset;
    uint16_t read_length;
    uint8_t* buffer_cur;
    uint8_t finished = 0;
    do
    {
        /* determine byte count to read at once */
        block_offset = offset & 0x01ff;
        read_length = 512 - block_offset;
        
        /* send single block request */
        if(sd_raw_send_command(CMD_READ_SINGLE_BLOCK, offset - block_offset))
        {
            unselect_card();
            return 0;
        }

        /* wait for data block (start byte 0xfe) */
        while(sd_raw_rec_byte() != 0xfe);

        /* read up to the data of interest */
        for(uint16_t i = 0; i < block_offset; ++i)
            sd_raw_rec_byte();

        /* read interval bytes of data and execute the callback */
        do
        {
            if(read_length < interval || length < interval)
                break;

            buffer_cur = buffer;
            for(uint16_t i = 0; i < interval; ++i)
                *buffer_cur++ = sd_raw_rec_byte();

            if(!callback(buffer, offset + (512 - read_length), p))
            {
                finished = 1;
                break;
            }

            read_length -= interval;
            length -= interval;

        } while(read_length > 0 && length > 0);
        
        /* read rest of data block */
        while(read_length-- > 0)
            sd_raw_rec_byte();
        
        /* read crc16 */
        sd_raw_rec_byte();
        sd_raw_rec_byte();

        if(length < interval)
            break;

        offset = offset - block_offset + 512;

    } while(!finished);
    
    /* deaddress card */
    unselect_card();

    /* let card some time to finish */
    sd_raw_rec_byte();

    return 1;
#endif
}

#if DOXYGEN || SD_RAW_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t sd_raw_write(offset_t offset, const uint8_t* buffer, uintptr_t length)
{
    if(sd_raw_locked())
        return 0;

    offset_t block_address;
    uint16_t block_offset;
    uint16_t write_length;
    while(length > 0)
    {
        /* determine byte count to write at once */
        block_offset = offset & 0x01ff;
        block_address = offset - block_offset;
        write_length = 512 - block_offset; /* write up to block border */
        if(write_length > length)
            write_length = length;
        
        /* Merge the data to write with the content of the block.
         * Use the cached block if available.
         */
        if(block_address != raw_block_address)
        {
#if SD_RAW_WRITE_BUFFERING
            if(!sd_raw_sync())
                return 0;
#endif

            if(block_offset || write_length < 512)
            {
                if(!sd_raw_read(block_address, raw_block, sizeof(raw_block)))
                    return 0;
            }
            raw_block_address = block_address;
        }

        if(buffer != raw_block)
        {
            memcpy(raw_block + block_offset, buffer, write_length);

#if SD_RAW_WRITE_BUFFERING
            raw_block_written = 0;

            if(length == write_length)
                return 1;
#endif
        }

        /* address card */
        select_card();

        /* send single block request */
        if(sd_raw_send_command(CMD_WRITE_SINGLE_BLOCK, block_address))
        {
            unselect_card();
            return 0;
        }

        /* send start byte */
        sd_raw_send_byte(0xfe);

        /* write byte block */
        uint8_t* cache = raw_block;
        for(uint16_t i = 0; i < 512; ++i)
            sd_raw_send_byte(*cache++);

        /* write dummy crc16 */
        sd_raw_send_byte(0xff);
        sd_raw_send_byte(0xff);

        /* wait while card is busy */
        while(sd_raw_rec_byte() != 0xff);
        sd_raw_rec_byte();

        /* deaddress card */
        unselect_card();

        buffer += write_length;
        offset += write_length;
        length -= write_length;

#if SD_RAW_WRITE_BUFFERING
        raw_block_written = 1;
#endif
    }

    return 1;
}
#endif

#if DOXYGEN || SD_RAW_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t sd_raw_write_interval(offset_t offset, uint8_t* buffer, uintptr_t length, sd_raw_write_interval_handler_t callback, void* p)
{
#if SD_RAW_SAVE_RAM
    #error "SD_RAW_WRITE_SUPPORT is not supported together with SD_RAW_SAVE_RAM"
#endif

    if(!buffer || !callback)
        return 0;

    uint8_t endless = (length == 0);
    while(endless || length > 0)
    {
        uint16_t bytes_to_write = callback(buffer, offset, p);
        if(!bytes_to_write)
            break;
        if(!endless && bytes_to_write > length)
            return 0;

        /* as writing is always buffered, we directly
         * hand over the request to sd_raw_write()
         */
        if(!sd_raw_write(offset, buffer, bytes_to_write))
            return 0;

        offset += bytes_to_write;
        length -= bytes_to_write;
    }

    return 1;
}
#endif

#if DOXYGEN || SD_RAW_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t sd_raw_sync()
{
#if SD_RAW_WRITE_BUFFERING
    if(raw_block_written)
        return 1;
    if(!sd_raw_write(raw_block_address, raw_block, sizeof(raw_block)))
        return 0;
    raw_block_written = 1;
#endif
    return 1;
}
#endif

//----------------------------------------------------------------------------
#if USE_MK3_SD_GET_INFO==1
uint8_t sd_raw_get_info(struct sd_raw_info* info)
{
    if(!info || !sd_raw_available())
        return 0;

    memset(info, 0, sizeof(*info));

    select_card();

    /* read cid register */
    if(sd_raw_send_command(CMD_SEND_CID, 0))
    {
        unselect_card();
        return 0;
    }
    while(sd_raw_rec_byte() != 0xfe);
    for(uint8_t i = 0; i < 18; ++i)
    {
        uint8_t b = sd_raw_rec_byte();

        switch(i)
        {
            case 0:
                info->manufacturer = b;
                break;
            case 1:
            case 2:
                info->oem[i - 1] = b;
                break;
            case 3:
            case 4:
            case 5:
            case 6:
            case 7:
                info->product[i - 3] = b;
                break;
            case 8:
                info->revision = b;
                break;
            case 9:
            case 10:
            case 11:
            case 12:
                info->serial |= (uint32_t) b << ((12 - i) * 8);
                break;
            case 13:
                info->manufacturing_year = b << 4;
                break;
            case 14:
                info->manufacturing_year |= b >> 4;
                info->manufacturing_month = b & 0x0f;
                break;
        }
    }

    /* read csd register */
    uint8_t csd_read_bl_len = 0;
    uint8_t csd_c_size_mult = 0;

    uint32_t csd_c_size = 0;

    uint8_t csd_structure = 0;
    if(sd_raw_send_command(CMD_SEND_CSD, 0))
    {
        unselect_card();
        return 0;
    }
    while(sd_raw_rec_byte() != 0xfe);
    for(uint8_t i = 0; i < 18; ++i)
    {
        uint8_t b = sd_raw_rec_byte();

        if(i == 0)
        {
            csd_structure = b >> 6;
        }
        else if(i == 14)
        {
            if(b & 0x40)
                info->flag_copy = 1;
            if(b & 0x20)
                info->flag_write_protect = 1;
            if(b & 0x10)
                info->flag_write_protect_temp = 1;
            info->format = (b & 0x0c) >> 2;
        }
        else
        {

            {
                switch(i)
                {
                    case 5:
                        csd_read_bl_len = b & 0x0f;
                        break;
                    case 6:
                        csd_c_size = b & 0x03;
                        csd_c_size <<= 8;
                        break;
                    case 7:
                        csd_c_size |= b;
                        csd_c_size <<= 2;
                        break;
                    case 8:
                        csd_c_size |= b >> 6;
                        ++csd_c_size;
                        break;
                    case 9:
                        csd_c_size_mult = b & 0x03;
                        csd_c_size_mult <<= 1;
                        break;
                    case 10:
                        csd_c_size_mult |= b >> 7;

                        info->capacity = (uint32_t) csd_c_size << (csd_c_size_mult + csd_read_bl_len + 2);
                        break;
                }
            }
        }
    }

    unselect_card();

    return 1;
}
#endif

//----------------------------------------------------------------------------
// SD_RAW.c [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// PARTITION.c [START]
//----------------------------------------------------------------------------

static struct partition_struct partition_handles[PARTITION_COUNT];

//----------------------------------------------------------------------------
struct partition_struct* partition_open(device_read_t device_read, device_read_interval_t device_read_interval, device_write_t device_write, device_write_interval_t device_write_interval, int8_t index)
{
    struct partition_struct* new_partition = 0;
    uint8_t buffer[0x10];

    if(!device_read || !device_read_interval || index >= 4)
        return 0;

    if(index >= 0)
    {
        /* read specified partition table index */
        if(!device_read(0x01be + index * 0x10, buffer, sizeof(buffer)))
            return 0;

        /* abort on empty partition entry */
        if(buffer[4] == 0x00)
            return 0;
    }

    /* allocate partition descriptor */
    new_partition = partition_handles;
    uint8_t i;
    for(i = 0; i < PARTITION_COUNT; ++i)
    {
        if(new_partition->type == PARTITION_TYPE_FREE)
            break;

        ++new_partition;
    }
    if(i >= PARTITION_COUNT)
        return 0;

    memset(new_partition, 0, sizeof(*new_partition));

    /* fill partition descriptor */
    new_partition->device_read = device_read;
    new_partition->device_read_interval = device_read_interval;
    new_partition->device_write = device_write;
    new_partition->device_write_interval = device_write_interval;

    if(index >= 0)
    {
        new_partition->type = buffer[4];
        new_partition->offset = read32(&buffer[8]);
        new_partition->length = read32(&buffer[12]);
    }
    else
    {
        new_partition->type = 0xff;
    }

    return new_partition;
}

//----------------------------------------------------------------------------
uint8_t partition_close(struct partition_struct* partition)
{
    if(!partition)
        return 0;

    /* destroy partition descriptor */
    partition->type = PARTITION_TYPE_FREE;

    return 1;
}

//----------------------------------------------------------------------------
// PARTITION.c [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// BYTEORDERING.c [START]
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
uint16_t read16(const uint8_t* p)
{
    return (((uint16_t) p[1]) << 8) |
           (((uint16_t) p[0]) << 0);
}

//----------------------------------------------------------------------------
uint32_t read32(const uint8_t* p)
{
    return (((uint32_t) p[3]) << 24) |
           (((uint32_t) p[2]) << 16) |
           (((uint32_t) p[1]) <<  8) |
           (((uint32_t) p[0]) <<  0);
}

//----------------------------------------------------------------------------
void write16(uint8_t* p, uint16_t i)
{
    p[1] = (uint8_t) ((i & 0xff00) >> 8);
    p[0] = (uint8_t) ((i & 0x00ff) >> 0);
}

//----------------------------------------------------------------------------
void write32(uint8_t* p, uint32_t i)
{
    p[3] = (uint8_t) ((i & 0xff000000) >> 24);
    p[2] = (uint8_t) ((i & 0x00ff0000) >> 16);
    p[1] = (uint8_t) ((i & 0x0000ff00) >>  8);
    p[0] = (uint8_t) ((i & 0x000000ff) >>  0);
}

//----------------------------------------------------------------------------
// BYTEORDERING.c [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// FAT.c [START]
//----------------------------------------------------------------------------

#define FAT16_CLUSTER_FREE 0x0000
#define FAT16_CLUSTER_RESERVED_MIN 0xfff0
#define FAT16_CLUSTER_RESERVED_MAX 0xfff6
#define FAT16_CLUSTER_BAD 0xfff7
#define FAT16_CLUSTER_LAST_MIN 0xfff8
#define FAT16_CLUSTER_LAST_MAX 0xffff

#define FAT32_CLUSTER_FREE 0x00000000
#define FAT32_CLUSTER_RESERVED_MIN 0x0ffffff0
#define FAT32_CLUSTER_RESERVED_MAX 0x0ffffff6
#define FAT32_CLUSTER_BAD 0x0ffffff7
#define FAT32_CLUSTER_LAST_MIN 0x0ffffff8
#define FAT32_CLUSTER_LAST_MAX 0x0fffffff

#define FAT_DIRENTRY_DELETED 0xe5
#define FAT_DIRENTRY_LFNLAST (1 << 6)
#define FAT_DIRENTRY_LFNSEQMASK ((1 << 6) - 1)


struct fat_header_struct
{
    offset_t size;

    offset_t fat_offset;
    uint32_t fat_size;

    uint16_t sector_size;
    uint16_t cluster_size;

    offset_t cluster_zero_offset;

    offset_t root_dir_offset;
};

struct fat_fs_struct
{
    struct partition_struct* partition;
    struct fat_header_struct header;
    cluster_t cluster_free;
};

struct fat_file_struct
{
    struct fat_fs_struct* fs;
    struct fat_dir_entry_struct dir_entry;
    offset_t pos;
    cluster_t pos_cluster;
};

struct fat_dir_struct
{
    struct fat_fs_struct* fs;
    struct fat_dir_entry_struct dir_entry;
    cluster_t entry_cluster;
    uint16_t entry_offset;
};

struct fat_read_dir_callback_arg
{
    struct fat_dir_entry_struct* dir_entry;
    uintptr_t bytes_read;
#if FAT_LFN_SUPPORT
    uint8_t checksum;
#endif
    uint8_t finished;
};

struct fat_usage_count_callback_arg
{
    cluster_t cluster_count;
    uintptr_t buffer_size;
};


static struct fat_fs_struct fat_fs_handles[FAT_FS_COUNT];
static struct fat_file_struct fat_file_handles[FAT_FILE_COUNT];
static struct fat_dir_struct fat_dir_handles[FAT_DIR_COUNT];


static uint8_t fat_read_header(struct fat_fs_struct* fs);
static cluster_t fat_get_next_cluster(const struct fat_fs_struct* fs, cluster_t cluster_num);
static offset_t fat_cluster_offset(const struct fat_fs_struct* fs, cluster_t cluster_num);
static uint8_t fat_dir_entry_read_callback(uint8_t* buffer, offset_t offset, void* p);
#if FAT_LFN_SUPPORT
static uint8_t fat_calc_83_checksum(const uint8_t* file_name_83);
#endif

#if USE_MK3_SD_GET_FS_FREE==1
static uint8_t fat_get_fs_free_16_callback(uint8_t* buffer, offset_t offset, void* p);
#endif

#if FAT_WRITE_SUPPORT
static cluster_t fat_append_clusters(struct fat_fs_struct* fs, cluster_t cluster_num, cluster_t count);
static uint8_t fat_free_clusters(struct fat_fs_struct* fs, cluster_t cluster_num);
static uint8_t fat_terminate_clusters(struct fat_fs_struct* fs, cluster_t cluster_num);
static uint8_t fat_clear_cluster(const struct fat_fs_struct* fs, cluster_t cluster_num);
static uintptr_t fat_clear_cluster_callback(uint8_t* buffer, offset_t offset, void* p);
static offset_t fat_find_offset_for_dir_entry(struct fat_fs_struct* fs, const struct fat_dir_struct* parent, const struct fat_dir_entry_struct* dir_entry);
static uint8_t fat_write_dir_entry(const struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry);
#endif

//----------------------------------------------------------------------------
struct fat_fs_struct* fat_open(struct partition_struct* partition)
{
    if(!partition ||
#if FAT_WRITE_SUPPORT
       !partition->device_write ||
       !partition->device_write_interval
#else
       0
#endif
      )
        return 0;

    struct fat_fs_struct* fs = fat_fs_handles;
    uint8_t i;
    for(i = 0; i < FAT_FS_COUNT; ++i)
    {
        if(!fs->partition)
            break;

        ++fs;
    }
    if(i >= FAT_FS_COUNT)
        return 0;


    memset(fs, 0, sizeof(*fs));

    fs->partition = partition;
    if(!fat_read_header(fs))
    {
        fs->partition = 0;

        return 0;
    }
    
    return fs;
}

//----------------------------------------------------------------------------
void fat_close(struct fat_fs_struct* fs)
{
    if(!fs)
        return;

    fs->partition = 0;
}

//----------------------------------------------------------------------------
uint8_t fat_read_header(struct fat_fs_struct* fs)
{
    if(!fs)
        return 0;

    struct partition_struct* partition = fs->partition;
    if(!partition)
        return 0;

    /* read fat parameters */
    uint8_t buffer[25];

    offset_t partition_offset = (offset_t) partition->offset * 512;
    if(!partition->device_read(partition_offset + 0x0b, buffer, sizeof(buffer)))
        return 0;

    uint16_t bytes_per_sector = read16(&buffer[0x00]);
    uint16_t reserved_sectors = read16(&buffer[0x03]);
    uint8_t sectors_per_cluster = buffer[0x02];
    uint8_t fat_copies = buffer[0x05];
    uint16_t max_root_entries = read16(&buffer[0x06]);
    uint16_t sector_count_16 = read16(&buffer[0x08]);
    uint16_t sectors_per_fat = read16(&buffer[0x0b]);
    uint32_t sector_count = read32(&buffer[0x15]);

    if(sector_count == 0)
    {
        if(sector_count_16 == 0)
            /* illegal volume size */
            return 0;
        else
            sector_count = sector_count_16;
    }

    if(sectors_per_fat == 0)
        /* this is not a FAT16 */
        return 0;

    /* determine the type of FAT we have here */
    uint32_t data_sector_count = sector_count
                                 - reserved_sectors

                                 - (uint32_t) sectors_per_fat * fat_copies

                                 - ((max_root_entries * 32 + bytes_per_sector - 1) / bytes_per_sector);
    uint32_t data_cluster_count = data_sector_count / sectors_per_cluster;
    if(data_cluster_count < 4085)
        /* this is a FAT12, not supported */
        return 0;
    else if(data_cluster_count < 65525)
        /* this is a FAT16 */
        partition->type = PARTITION_TYPE_FAT16;
    else
        /* this is a FAT32 */
        partition->type = PARTITION_TYPE_FAT32;

    /* fill header information */
    struct fat_header_struct* header = &fs->header;
    memset(header, 0, sizeof(*header));
    
    header->size = (offset_t) sector_count * bytes_per_sector;

    header->fat_offset = /* jump to partition */
                         partition_offset +
                         /* jump to fat */
                         (offset_t) reserved_sectors * bytes_per_sector;
    header->fat_size = (data_cluster_count + 2) * (partition->type == PARTITION_TYPE_FAT16 ? 2 : 4);

    header->sector_size = bytes_per_sector;
    header->cluster_size = (uint16_t) bytes_per_sector * sectors_per_cluster;


    {
        header->root_dir_offset = /* jump to fats */
                                  header->fat_offset +
                                  /* jump to root directory entries */
                                  (offset_t) fat_copies * sectors_per_fat * bytes_per_sector;

        header->cluster_zero_offset = /* jump to root directory entries */
                                      header->root_dir_offset +
                                      /* skip root directory entries */
                                      (offset_t) max_root_entries * 32;
    }

    return 1;
}

//----------------------------------------------------------------------------
cluster_t fat_get_next_cluster(const struct fat_fs_struct* fs, cluster_t cluster_num)
{
    if(!fs || cluster_num < 2)
        return 0;

    {
        /* read appropriate fat entry */
        uint16_t fat_entry;
        if(!fs->partition->device_read(fs->header.fat_offset + (offset_t) cluster_num * sizeof(fat_entry), (uint8_t*) &fat_entry, sizeof(fat_entry)))
            return 0;

        /* determine next cluster from fat */
        cluster_num = ltoh16(fat_entry);
        
        if(cluster_num == FAT16_CLUSTER_FREE ||
           cluster_num == FAT16_CLUSTER_BAD ||
           (cluster_num >= FAT16_CLUSTER_RESERVED_MIN && cluster_num <= FAT16_CLUSTER_RESERVED_MAX) ||
           (cluster_num >= FAT16_CLUSTER_LAST_MIN && cluster_num <= FAT16_CLUSTER_LAST_MAX))
            return 0;
    }

    return cluster_num;
}

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
cluster_t fat_append_clusters(struct fat_fs_struct* fs, cluster_t cluster_num, cluster_t count)
{
    if(!fs)
        return 0;

    device_read_t device_read = fs->partition->device_read;
    device_write_t device_write = fs->partition->device_write;
    offset_t fat_offset = fs->header.fat_offset;
    cluster_t count_left = count;
    cluster_t cluster_current = fs->cluster_free;
    cluster_t cluster_next = 0;
    cluster_t cluster_count;
    uint16_t fat_entry16;

    cluster_count = fs->header.fat_size / sizeof(fat_entry16);

    fs->cluster_free = 0;
    for(cluster_t cluster_left = cluster_count; cluster_left > 0; --cluster_left, ++cluster_current)
    {
        if(cluster_current < 2 || cluster_current >= cluster_count)
            cluster_current = 2;

        {
            if(!device_read(fat_offset + (offset_t) cluster_current * sizeof(fat_entry16), (uint8_t*) &fat_entry16, sizeof(fat_entry16)))
                return 0;
        }


        {
            /* check if this is a free cluster */
            if(fat_entry16 != HTOL16(FAT16_CLUSTER_FREE))
                continue;

            /* If we don't need this free cluster for the
             * current allocation, we keep it in mind for
             * the next time.
             */
            if(count_left == 0)
            {
                fs->cluster_free = cluster_current;
                break;
            }

            /* allocate cluster */
            if(cluster_next == 0)
                fat_entry16 = HTOL16(FAT16_CLUSTER_LAST_MAX);
            else
                fat_entry16 = htol16((uint16_t) cluster_next);

            if(!device_write(fat_offset + (offset_t) cluster_current * sizeof(fat_entry16), (uint8_t*) &fat_entry16, sizeof(fat_entry16)))
                break;
        }

        cluster_next = cluster_current;
        --count_left;
    }

    do
    {
        if(count_left > 0)
            break;

        /* We allocated a new cluster chain. Now join
         * it with the existing one (if any).
         */
        if(cluster_num >= 2)
        {

            {
                fat_entry16 = htol16((uint16_t) cluster_next);

                if(!device_write(fat_offset + (offset_t) cluster_num * sizeof(fat_entry16), (uint8_t*) &fat_entry16, sizeof(fat_entry16)))
                    break;
            }
        }

        return cluster_next;

    } while(0);

    /* No space left on device or writing error.
     * Free up all clusters already allocated.
     */
    fat_free_clusters(fs, cluster_next);

    return 0;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_free_clusters(struct fat_fs_struct* fs, cluster_t cluster_num)
{
    if(!fs || cluster_num < 2)
        return 0;

    offset_t fat_offset = fs->header.fat_offset;

    {
        uint16_t fat_entry;
        while(cluster_num)
        {
            if(!fs->partition->device_read(fat_offset + (offset_t) cluster_num * sizeof(fat_entry), (uint8_t*) &fat_entry, sizeof(fat_entry)))
                return 0;

            /* get next cluster of current cluster before freeing current cluster */
            uint16_t cluster_num_next = ltoh16(fat_entry);

            if(cluster_num_next == FAT16_CLUSTER_FREE)
                return 1;
            if(cluster_num_next == FAT16_CLUSTER_BAD ||
               (cluster_num_next >= FAT16_CLUSTER_RESERVED_MIN &&
                cluster_num_next <= FAT16_CLUSTER_RESERVED_MAX
               )
              )
                return 0;
            if(cluster_num_next >= FAT16_CLUSTER_LAST_MIN && cluster_num_next <= FAT16_CLUSTER_LAST_MAX)
                cluster_num_next = 0;

            /* free cluster */
            fat_entry = HTOL16(FAT16_CLUSTER_FREE);
            fs->partition->device_write(fat_offset + (offset_t) cluster_num * sizeof(fat_entry), (uint8_t*) &fat_entry, sizeof(fat_entry));

            /* We continue in any case here, even if freeing the cluster failed.
             * The cluster is lost, but maybe we can still free up some later ones.
             */

            cluster_num = cluster_num_next;
        }
    }

    return 1;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_terminate_clusters(struct fat_fs_struct* fs, cluster_t cluster_num)
{
    if(!fs || cluster_num < 2)
        return 0;

    /* fetch next cluster before overwriting the cluster entry */
    cluster_t cluster_num_next = fat_get_next_cluster(fs, cluster_num);

    /* mark cluster as the last one */
    {
        uint16_t fat_entry = HTOL16(FAT16_CLUSTER_LAST_MAX);
        if(!fs->partition->device_write(fs->header.fat_offset + (offset_t) cluster_num * sizeof(fat_entry), (uint8_t*) &fat_entry, sizeof(fat_entry)))
            return 0;
    }

    /* free remaining clusters */
    if(cluster_num_next)
        return fat_free_clusters(fs, cluster_num_next);
    else
        return 1;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_clear_cluster(const struct fat_fs_struct* fs, cluster_t cluster_num)
{
    if(cluster_num < 2)
        return 0;

    offset_t cluster_offset = fat_cluster_offset(fs, cluster_num);

    uint8_t zero[16];
    memset(zero, 0, sizeof(zero));
    return fs->partition->device_write_interval(cluster_offset,
                                                zero,
                                                fs->header.cluster_size,
                                                fat_clear_cluster_callback,
                                                0
                                               );
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uintptr_t fat_clear_cluster_callback(uint8_t* buffer, offset_t offset, void* p)
{
    return 16;
}
#endif

//----------------------------------------------------------------------------
offset_t fat_cluster_offset(const struct fat_fs_struct* fs, cluster_t cluster_num)
{
    if(!fs || cluster_num < 2)
        return 0;

    return fs->header.cluster_zero_offset + (offset_t) (cluster_num - 2) * fs->header.cluster_size;
}

//----------------------------------------------------------------------------
uint8_t fat_get_dir_entry_of_path(struct fat_fs_struct* fs, const char* path, struct fat_dir_entry_struct* dir_entry)
{
    if(!fs || !path || path[0] == '\0' || !dir_entry)
        return 0;

    if(path[0] == '/')
        ++path;

    /* begin with the root directory */
    memset(dir_entry, 0, sizeof(*dir_entry));
    dir_entry->attributes = FAT_ATTRIB_DIR;

    while(1)
    {
        if(path[0] == '\0')
            return 1;

        struct fat_dir_struct* dd = fat_open_dir(fs, dir_entry);
        if(!dd)
            break;

        /* extract the next hierarchy we will search for */
        const char* sub_path = strchr(path, '/');
        uint8_t length_to_sep;
        if(sub_path)
        {
            length_to_sep = sub_path - path;
            ++sub_path;
        }
        else
        {
            length_to_sep = strlen(path);
            sub_path = path + length_to_sep;
        }
        
        /* read directory entries */
        while(fat_read_dir(dd, dir_entry))
        {
            /* check if we have found the next hierarchy */
            if((strlen(dir_entry->long_name) != length_to_sep ||
                strncmp(path, dir_entry->long_name, length_to_sep) != 0))
                continue;

            fat_close_dir(dd);
            dd = 0;

            if(path[length_to_sep] == '\0')
                /* we iterated through the whole path and have found the file */
                return 1;

            if(dir_entry->attributes & FAT_ATTRIB_DIR)
            {
                /* we found a parent directory of the file we are searching for */
                path = sub_path;
                break;
            }

            /* a parent of the file exists, but not the file itself */
            return 0;
        }

        fat_close_dir(dd);
    }
    
    return 0;
}

//----------------------------------------------------------------------------
struct fat_file_struct* fat_open_file(struct fat_fs_struct* fs, const struct fat_dir_entry_struct* dir_entry)
{
    if(!fs || !dir_entry || (dir_entry->attributes & FAT_ATTRIB_DIR))
        return 0;

    struct fat_file_struct* fd = fat_file_handles;
    uint8_t i;
    for(i = 0; i < FAT_FILE_COUNT; ++i)
    {
        if(!fd->fs)
            break;

        ++fd;
    }
    if(i >= FAT_FILE_COUNT)
        return 0;
    
    memcpy(&fd->dir_entry, dir_entry, sizeof(*dir_entry));
    fd->fs = fs;
    fd->pos = 0;
    fd->pos_cluster = dir_entry->cluster;

    return fd;
}

//----------------------------------------------------------------------------
void fat_close_file(struct fat_file_struct* fd)
{
    if(fd)
    {
#if FAT_DELAY_DIRENTRY_UPDATE
        /* write directory entry */
        fat_write_dir_entry(fd->fs, &fd->dir_entry);
#endif

        fd->fs = 0;
    }
}

//----------------------------------------------------------------------------
intptr_t fat_read_file(struct fat_file_struct* fd, uint8_t* buffer, uintptr_t buffer_len)
{
    /* check arguments */
    if(!fd || !buffer || buffer_len < 1)
        return -1;

    /* determine number of bytes to read */
    if(fd->pos + buffer_len > fd->dir_entry.file_size)
        buffer_len = fd->dir_entry.file_size - fd->pos;
    if(buffer_len == 0)
        return 0;
    
    uint16_t cluster_size = fd->fs->header.cluster_size;
    cluster_t cluster_num = fd->pos_cluster;
    uintptr_t buffer_left = buffer_len;
    uint16_t first_cluster_offset = (uint16_t) (fd->pos & (cluster_size - 1));

    /* find cluster in which to start reading */
    if(!cluster_num)
    {
        cluster_num = fd->dir_entry.cluster;
        
        if(!cluster_num)
        {
            if(!fd->pos)
                return 0;
            else
                return -1;
        }

        if(fd->pos)
        {
            uint32_t pos = fd->pos;
            while(pos >= cluster_size)
            {
                pos -= cluster_size;
                cluster_num = fat_get_next_cluster(fd->fs, cluster_num);
                if(!cluster_num)
                    return -1;
            }
        }
    }
    
    /* read data */
    do
    {
        /* calculate data size to copy from cluster */
        offset_t cluster_offset = fat_cluster_offset(fd->fs, cluster_num) + first_cluster_offset;
        uint16_t copy_length = cluster_size - first_cluster_offset;
        if(copy_length > buffer_left)
            copy_length = buffer_left;

        /* read data */
        if(!fd->fs->partition->device_read(cluster_offset, buffer, copy_length))
            return buffer_len - buffer_left;

        /* calculate new file position */
        buffer += copy_length;
        buffer_left -= copy_length;
        fd->pos += copy_length;

        if(first_cluster_offset + copy_length >= cluster_size)
        {
            /* we are on a cluster boundary, so get the next cluster */
            if((cluster_num = fat_get_next_cluster(fd->fs, cluster_num)))
            {
                first_cluster_offset = 0;
            }
            else
            {
                fd->pos_cluster = 0;
                return buffer_len - buffer_left;
            }
        }

        fd->pos_cluster = cluster_num;

    } while(buffer_left > 0); /* check if we are done */

    return buffer_len;
}

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
intptr_t fat_write_file(struct fat_file_struct* fd, const uint8_t* buffer, uintptr_t buffer_len)
{
    /* check arguments */
    if(!fd || !buffer || buffer_len < 1)
        return -1;
    if(fd->pos > fd->dir_entry.file_size)
        return -1;

    uint16_t cluster_size = fd->fs->header.cluster_size;
    cluster_t cluster_num = fd->pos_cluster;
    uintptr_t buffer_left = buffer_len;
    uint16_t first_cluster_offset = (uint16_t) (fd->pos & (cluster_size - 1));

    /* find cluster in which to start writing */
    if(!cluster_num)
    {
        cluster_num = fd->dir_entry.cluster;
        
        if(!cluster_num)
        {
            if(!fd->pos)
            {
                /* empty file */
                fd->dir_entry.cluster = cluster_num = fat_append_clusters(fd->fs, 0, 1);
                if(!cluster_num)
                    return 0;
            }
            else
            {
                return -1;
            }
        }

        if(fd->pos)
        {
            uint32_t pos = fd->pos;
            cluster_t cluster_num_next;
            while(pos >= cluster_size)
            {
                pos -= cluster_size;
                cluster_num_next = fat_get_next_cluster(fd->fs, cluster_num);
                if(!cluster_num_next)
                {
                    if(pos != 0)
                        return -1; /* current file position points beyond end of file */

                    /* the file exactly ends on a cluster boundary, and we append to it */
                    cluster_num_next = fat_append_clusters(fd->fs, cluster_num, 1);
                    if(!cluster_num_next)
                        return 0;
                }

                cluster_num = cluster_num_next;
            }
        }
    }
    
    /* write data */
    do
    {
        /* calculate data size to write to cluster */
        offset_t cluster_offset = fat_cluster_offset(fd->fs, cluster_num) + first_cluster_offset;
        uint16_t write_length = cluster_size - first_cluster_offset;
        if(write_length > buffer_left)
            write_length = buffer_left;

        /* write data which fits into the current cluster */
        if(!fd->fs->partition->device_write(cluster_offset, buffer, write_length))
            break;

        /* calculate new file position */
        buffer += write_length;
        buffer_left -= write_length;
        fd->pos += write_length;

        if(first_cluster_offset + write_length >= cluster_size)
        {
            /* we are on a cluster boundary, so get the next cluster */
            cluster_t cluster_num_next = fat_get_next_cluster(fd->fs, cluster_num);
            if(!cluster_num_next && buffer_left > 0)
                /* we reached the last cluster, append a new one */
                cluster_num_next = fat_append_clusters(fd->fs, cluster_num, 1);
            if(!cluster_num_next)
            {
                fd->pos_cluster = 0;
                break;
            }

            cluster_num = cluster_num_next;
            first_cluster_offset = 0;
        }

        fd->pos_cluster = cluster_num;

    } while(buffer_left > 0); /* check if we are done */

    /* update directory entry */
    if(fd->pos > fd->dir_entry.file_size)
    {
#if !FAT_DELAY_DIRENTRY_UPDATE
        uint32_t size_old = fd->dir_entry.file_size;
#endif

        /* update file size */
        fd->dir_entry.file_size = fd->pos;

#if !FAT_DELAY_DIRENTRY_UPDATE
        /* write directory entry */
        if(!fat_write_dir_entry(fd->fs, &fd->dir_entry))
        {
            /* We do not return an error here since we actually wrote
             * some data to disk. So we calculate the amount of data
             * we wrote to disk and which lies within the old file size.
             */
            buffer_left = fd->pos - size_old;
            fd->pos = size_old;
        }
#endif
    }

    return buffer_len - buffer_left;
}
#endif

//----------------------------------------------------------------------------
#if USE_MK3_SD_SEEK_FILE==1
uint8_t fat_seek_file(struct fat_file_struct* fd, int32_t* offset, uint8_t whence)
{
    if(!fd || !offset)
        return 0;

    uint32_t new_pos = fd->pos;
    switch(whence)
    {
        case FAT_SEEK_SET:
            new_pos = *offset;
            break;
        case FAT_SEEK_CUR:
            new_pos += *offset;
            break;
        case FAT_SEEK_END:
            new_pos = fd->dir_entry.file_size + *offset;
            break;
        default:
            return 0;
    }

    if(new_pos > fd->dir_entry.file_size
#if FAT_WRITE_SUPPORT
       && !fat_resize_file(fd, new_pos)
#endif
       )
        return 0;

    fd->pos = new_pos;
    fd->pos_cluster = 0;

    *offset = (int32_t) new_pos;
    return 1;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_resize_file(struct fat_file_struct* fd, uint32_t size)
{
    if(!fd)
        return 0;

    cluster_t cluster_num = fd->dir_entry.cluster;
    uint16_t cluster_size = fd->fs->header.cluster_size;
    uint32_t size_new = size;

    do
    {
        if(cluster_num == 0 && size_new == 0)
            /* the file stays empty */
            break;

        /* seek to the next cluster as long as we need the space */
        while(size_new > cluster_size)
        {
            /* get next cluster of file */
            cluster_t cluster_num_next = fat_get_next_cluster(fd->fs, cluster_num);
            if(cluster_num_next)
            {
                cluster_num = cluster_num_next;
                size_new -= cluster_size;
            }
            else
            {
                break;
            }
        }

        if(size_new > cluster_size || cluster_num == 0)
        {
            /* Allocate new cluster chain and append
             * it to the existing one, if available.
             */
            cluster_t cluster_count = (size_new + cluster_size - 1) / cluster_size;
            cluster_t cluster_new_chain = fat_append_clusters(fd->fs, cluster_num, cluster_count);
            if(!cluster_new_chain)
                return 0;

            if(!cluster_num)
            {
                cluster_num = cluster_new_chain;
                fd->dir_entry.cluster = cluster_num;
            }
        }

        /* write new directory entry */
        fd->dir_entry.file_size = size;
        if(size == 0)
            fd->dir_entry.cluster = 0;
        if(!fat_write_dir_entry(fd->fs, &fd->dir_entry))
            return 0;

        if(size == 0)
        {
            /* free all clusters of file */
            fat_free_clusters(fd->fs, cluster_num);
        }
        else if(size_new <= cluster_size)
        {
            /* free all clusters no longer needed */
            fat_terminate_clusters(fd->fs, cluster_num);
        }

    } while(0);

    /* correct file position */
    if(size < fd->pos)
    {
        fd->pos = size;
        fd->pos_cluster = 0;
    }

    return 1;
}
#endif

//----------------------------------------------------------------------------
struct fat_dir_struct* fat_open_dir(struct fat_fs_struct* fs, const struct fat_dir_entry_struct* dir_entry)
{
    if(!fs || !dir_entry || !(dir_entry->attributes & FAT_ATTRIB_DIR))
        return 0;

    struct fat_dir_struct* dd = fat_dir_handles;
    uint8_t i;
    for(i = 0; i < FAT_DIR_COUNT; ++i)
    {
        if(!dd->fs)
            break;

        ++dd;
    }
    if(i >= FAT_DIR_COUNT)
        return 0;
    
    memcpy(&dd->dir_entry, dir_entry, sizeof(*dir_entry));
    dd->fs = fs;
    dd->entry_cluster = dir_entry->cluster;
    dd->entry_offset = 0;

    return dd;
}

//----------------------------------------------------------------------------
void fat_close_dir(struct fat_dir_struct* dd)
{
    if(dd)
        dd->fs = 0;

}

//----------------------------------------------------------------------------
uint8_t fat_read_dir(struct fat_dir_struct* dd, struct fat_dir_entry_struct* dir_entry)
{
    if(!dd || !dir_entry)
        return 0;

    /* get current position of directory handle */
    struct fat_fs_struct* fs = dd->fs;
    const struct fat_header_struct* header = &fs->header;
    uint16_t cluster_size = header->cluster_size;
    cluster_t cluster_num = dd->entry_cluster;
    uint16_t cluster_offset = dd->entry_offset;
    struct fat_read_dir_callback_arg arg;

    if(cluster_offset >= cluster_size)
    {
        /* The latest call hit the border of the last cluster in
         * the chain, but it still returned a directory entry.
         * So we now reset the handle and signal the caller the
         * end of the listing.
         */
        fat_reset_dir(dd);
        return 0;
    }

    /* reset callback arguments */
    memset(&arg, 0, sizeof(arg));
    memset(dir_entry, 0, sizeof(*dir_entry));
    arg.dir_entry = dir_entry;

    /* check if we read from the root directory */
    if(cluster_num == 0)
    {
            cluster_size = header->cluster_zero_offset - header->root_dir_offset;
    }

    /* read entries */
    uint8_t buffer[32];
    while(!arg.finished)
    {
        /* read directory entries up to the cluster border */
        uint16_t cluster_left = cluster_size - cluster_offset;
        offset_t pos = cluster_offset;
        if(cluster_num == 0)
            pos += header->root_dir_offset;
        else
            pos += fat_cluster_offset(fs, cluster_num);

        arg.bytes_read = 0;
        if(!fs->partition->device_read_interval(pos,
                                                buffer,
                                                sizeof(buffer),
                                                cluster_left,
                                                fat_dir_entry_read_callback,
                                                &arg)
          )
            return 0;

        cluster_offset += arg.bytes_read;

        if(cluster_offset >= cluster_size)
        {
            /* we reached the cluster border and switch to the next cluster */

            /* get number of next cluster */
            if((cluster_num = fat_get_next_cluster(fs, cluster_num)) != 0)
            {
                cluster_offset = 0;
                continue;
            }

            /* we are at the end of the cluster chain */
            if(!arg.finished)
            {
                /* directory entry not found, reset directory handle */
                fat_reset_dir(dd);
                return 0;
            }
            else
            {
                /* The current execution of the function has been successful,
                 * so we can not signal an end of the directory listing to
                 * the caller, but must wait for the next call. So we keep an
                 * invalid cluster offset to mark this directory handle's
                 * traversal as finished.
                 */
            }

            break;
        }
    }

    dd->entry_cluster = cluster_num;
    dd->entry_offset = cluster_offset;

    return arg.finished;
}


//----------------------------------------------------------------------------
uint8_t fat_reset_dir(struct fat_dir_struct* dd)
{
    if(!dd)
        return 0;

    dd->entry_cluster = dd->dir_entry.cluster;
    dd->entry_offset = 0;
    return 1;
}

//----------------------------------------------------------------------------
uint8_t fat_dir_entry_read_callback(uint8_t* buffer, offset_t offset, void* p)
{
    struct fat_read_dir_callback_arg* arg = (struct fat_read_dir_callback_arg*) p;	
    struct fat_dir_entry_struct* dir_entry = arg->dir_entry;

    arg->bytes_read += 32;

    /* skip deleted or empty entries */
    if(buffer[0] == FAT_DIRENTRY_DELETED || !buffer[0])
    {
#if FAT_LFN_SUPPORT
        arg->checksum = 0;
#endif
        return 1;
    }

#if !FAT_LFN_SUPPORT
    /* skip lfn entries */
    if(buffer[11] == 0x0f)
        return 1;
#endif

    char* long_name = dir_entry->long_name;
#if FAT_LFN_SUPPORT
    if(buffer[11] == 0x0f)
    {
        /* checksum validation */
        if(arg->checksum == 0 || arg->checksum != buffer[13])
        {
            /* reset directory entry */
            memset(dir_entry, 0, sizeof(*dir_entry));

            arg->checksum = buffer[13];
            dir_entry->entry_offset = offset;
        }

        /* lfn supports unicode, but we do not, for now.
         * So we assume pure ascii and read only every
         * second byte.
         */
        uint16_t char_offset = ((buffer[0] & 0x3f) - 1) * 13;
        const uint8_t char_mapping[] = { 1, 3, 5, 7, 9, 14, 16, 18, 20, 22, 24, 28, 30 };
        for(uint8_t i = 0; i <= 12 && char_offset + i < sizeof(dir_entry->long_name) - 1; ++i)
            long_name[char_offset + i] = buffer[char_mapping[i]];

        return 1;
    }
    else
#endif
    {
#if FAT_LFN_SUPPORT
        /* if we do not have a long name or the previous lfn does not match, take the 8.3 name */
        if(long_name[0] == '\0' || arg->checksum != fat_calc_83_checksum(buffer))
#endif
        {
            /* reset directory entry */
            memset(dir_entry, 0, sizeof(*dir_entry));
            dir_entry->entry_offset = offset;

            uint8_t i;
            for(i = 0; i < 8; ++i)
            {
                if(buffer[i] == ' ')
                    break;
                long_name[i] = buffer[i];

                /* Windows NT and later versions do not store lfn entries
                 * for 8.3 names which have a lowercase basename, extension
                 * or both when everything else is uppercase. They use two
                 * extra bits to signal a lowercase basename or extension.
                 */
                if((buffer[12] & 0x08) && buffer[i] >= 'A' && buffer[i] <= 'Z')
                    long_name[i] += 'a' - 'A';
            }
            if(long_name[0] == 0x05)
                long_name[0] = (char) FAT_DIRENTRY_DELETED;

            if(buffer[8] != ' ')
            {
                long_name[i++] = '.';

                uint8_t j = 8;
                for(; j < 11; ++j)
                {
                    if(buffer[j] == ' ')
                        break;
                    long_name[i] = buffer[j];

                    /* See above for the lowercase 8.3 name handling of
                     * Windows NT and later.
                     */
                    if((buffer[12] & 0x10) && buffer[j] >= 'A' && buffer[j] <= 'Z')
                        long_name[i] += 'a' - 'A';

                    ++i;
                }
            } 

            long_name[i] = '\0';
        }
        
        /* extract properties of file and store them within the structure */
        dir_entry->attributes = buffer[11];
        dir_entry->cluster = read16(&buffer[26]);

        dir_entry->file_size = read32(&buffer[28]);

        arg->finished = 1;
        return 0;
    }
}

#if DOXYGEN || FAT_LFN_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_calc_83_checksum(const uint8_t* file_name_83)
{
    uint8_t checksum = file_name_83[0];
    for(uint8_t i = 1; i < 11; ++i)
        checksum = ((checksum >> 1) | (checksum << 7)) + file_name_83[i];

    return checksum;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
offset_t fat_find_offset_for_dir_entry(struct fat_fs_struct* fs, const struct fat_dir_struct* parent, const struct fat_dir_entry_struct* dir_entry)
{
    if(!fs || !dir_entry)
        return 0;

    /* search for a place where to write the directory entry to disk */
#if FAT_LFN_SUPPORT
    uint8_t free_dir_entries_needed = (strlen(dir_entry->long_name) + 12) / 13 + 1;
    uint8_t free_dir_entries_found = 0;
#endif
    cluster_t cluster_num = parent->dir_entry.cluster;
    offset_t dir_entry_offset = 0;
    offset_t offset = 0;
    offset_t offset_to = 0;

    if(cluster_num == 0)
    {

        {
            /* we read/write from the root directory entry */
            offset = fs->header.root_dir_offset;
            offset_to = fs->header.cluster_zero_offset;
            dir_entry_offset = offset;
        }
    }
    
    while(1)
    {
        if(offset == offset_to)
        {
            if(cluster_num == 0)
                /* We iterated through the whole root directory and
                 * could not find enough space for the directory entry.
                 */
                return 0;

            if(offset)
            {
                /* We reached a cluster boundary and have to
                 * switch to the next cluster.
                 */

                cluster_t cluster_next = fat_get_next_cluster(fs, cluster_num);
                if(!cluster_next)
                {
                    cluster_next = fat_append_clusters(fs, cluster_num, 1);
                    if(!cluster_next)
                        return 0;

                    /* we appended a new cluster and know it is free */
                    dir_entry_offset = fs->header.cluster_zero_offset +
                                       (offset_t) (cluster_next - 2) * fs->header.cluster_size;

                    /* clear cluster to avoid garbage directory entries */
                    fat_clear_cluster(fs, cluster_next);

                    break;
                }
                cluster_num = cluster_next;
            }

            offset = fat_cluster_offset(fs, cluster_num);
            offset_to = offset + fs->header.cluster_size;
            dir_entry_offset = offset;
#if FAT_LFN_SUPPORT
            free_dir_entries_found = 0;
#endif
        }
        
        /* read next lfn or 8.3 entry */
        uint8_t first_char;
        if(!fs->partition->device_read(offset, &first_char, sizeof(first_char)))
            return 0;

        /* check if we found a free directory entry */
        if(first_char == FAT_DIRENTRY_DELETED || !first_char)
        {
            /* check if we have the needed number of available entries */
#if FAT_LFN_SUPPORT
            ++free_dir_entries_found;
            if(free_dir_entries_found >= free_dir_entries_needed)
#endif
                break;

            offset += 32;
        }
        else
        {
            offset += 32;
            dir_entry_offset = offset;
#if FAT_LFN_SUPPORT
            free_dir_entries_found = 0;
#endif
        }
    }

    return dir_entry_offset;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_write_dir_entry(const struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry)
{
    if(!fs || !dir_entry)
        return 0;

    device_write_t device_write = fs->partition->device_write;
    offset_t offset = dir_entry->entry_offset;
    const char* name = dir_entry->long_name;
    uint8_t name_len = strlen(name);
#if FAT_LFN_SUPPORT
    uint8_t lfn_entry_count = (name_len + 12) / 13;
#endif
    uint8_t buffer[32];

    /* write 8.3 entry */

    /* generate 8.3 file name */
    memset(&buffer[0], ' ', 11);
    char* name_ext = strrchr(name, '.');
    if(name_ext && *++name_ext)
    {
        uint8_t name_ext_len = strlen(name_ext);
        name_len -= name_ext_len + 1;

        if(name_ext_len > 3)
#if FAT_LFN_SUPPORT
            name_ext_len = 3;
#else
            return 0;
#endif
        
        memcpy(&buffer[8], name_ext, name_ext_len);
    }
    
    if(name_len <= 8)
    {
        memcpy(buffer, name, name_len);

#if FAT_LFN_SUPPORT
        /* For now, we create lfn entries for all files,
         * except the "." and ".." directory references.
         * This is to avoid difficulties with capitalization,
         * as 8.3 filenames allow uppercase letters only.
         *
         * Theoretically it would be possible to leave
         * the 8.3 entry alone if the basename and the
         * extension have no mixed capitalization.
         */
        if(name[0] == '.' &&
           ((name[1] == '.' && name[2] == '\0') ||
            name[1] == '\0')
          )
            lfn_entry_count = 0;
#endif
    }
    else
    {
#if FAT_LFN_SUPPORT
        memcpy(buffer, name, 8);

        /* Minimize 8.3 name clashes by appending
         * the lower byte of the cluster number.
         */
        uint8_t num = dir_entry->cluster & 0xff;

        buffer[6] = (num < 0xa0) ? ('0' + (num >> 4)) : ('a' + (num >> 4));
        num &= 0x0f;
        buffer[7] = (num < 0x0a) ? ('0' + num) : ('a' + num);
#else
        return 0;
#endif
    }
    if(buffer[0] == FAT_DIRENTRY_DELETED)
        buffer[0] = 0x05;

    /* fill directory entry buffer */
    memset(&buffer[11], 0, sizeof(buffer) - 11);
    buffer[0x0b] = dir_entry->attributes;

    write16(&buffer[0x1a], dir_entry->cluster);
    write32(&buffer[0x1c], dir_entry->file_size);

    /* write to disk */
#if FAT_LFN_SUPPORT
    if(!device_write(offset + (uint16_t) lfn_entry_count * 32, buffer, sizeof(buffer)))
#else
    if(!device_write(offset, buffer, sizeof(buffer)))
#endif
        return 0;
    
#if FAT_LFN_SUPPORT
    /* calculate checksum of 8.3 name */
    uint8_t checksum = fat_calc_83_checksum(buffer);
    
    /* write lfn entries */
    for(uint8_t lfn_entry = lfn_entry_count; lfn_entry > 0; --lfn_entry)
    {
        memset(buffer, 0xff, sizeof(buffer));
        
        /* set file name */
        const char* long_name_curr = name + (lfn_entry - 1) * 13;
        uint8_t i = 1;
        while(i < 0x1f)
        {
            buffer[i++] = *long_name_curr;
            buffer[i++] = 0;

            switch(i)
            {
                case 0x0b:
                    i = 0x0e;
                    break;
                case 0x1a:
                    i = 0x1c;
                    break;
            }

            if(!*long_name_curr++)
                break;
        }
        
        /* set index of lfn entry */
        buffer[0x00] = lfn_entry;
        if(lfn_entry == lfn_entry_count)
            buffer[0x00] |= FAT_DIRENTRY_LFNLAST;

        /* mark as lfn entry */
        buffer[0x0b] = 0x0f;

        /* set 8.3 checksum */
        buffer[0x0d] = checksum;

        /* clear reserved bytes */
        buffer[0x0c] = 0;
        buffer[0x1a] = 0;
        buffer[0x1b] = 0;

        /* write entry */
        device_write(offset, buffer, sizeof(buffer));
    
        offset += sizeof(buffer);
    }
#endif
    
    return 1;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_create_file(struct fat_dir_struct* parent, const char* file, struct fat_dir_entry_struct* dir_entry)
{
    if(!parent || !file || !file[0] || !dir_entry)
        return 0;

    /* check if the file already exists */
    while(1)
    {
        if(!fat_read_dir(parent, dir_entry))
            break;

        if(strcmp(file, dir_entry->long_name) == 0)
        {
            fat_reset_dir(parent);
            return 2;
        }
    }

    struct fat_fs_struct* fs = parent->fs;

    /* prepare directory entry with values already known */
    memset(dir_entry, 0, sizeof(*dir_entry));
    strncpy(dir_entry->long_name, file, sizeof(dir_entry->long_name) - 1);

    /* find place where to store directory entry */
    if(!(dir_entry->entry_offset = fat_find_offset_for_dir_entry(fs, parent, dir_entry)))
        return 0;
    
    /* write directory entry to disk */
    if(!fat_write_dir_entry(fs, dir_entry))
        return 0;
    
    return 1;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_delete_file(struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry)
{
    if(!fs || !dir_entry)
        return 0;

    /* get offset of the file's directory entry */
    offset_t dir_entry_offset = dir_entry->entry_offset;
    if(!dir_entry_offset)
        return 0;

#if FAT_LFN_SUPPORT
    uint8_t buffer[12];
    while(1)
    {
        /* read directory entry */
        if(!fs->partition->device_read(dir_entry_offset, buffer, sizeof(buffer)))
            return 0;
        
        /* mark the directory entry as deleted */
        buffer[0] = FAT_DIRENTRY_DELETED;
        
        /* write back entry */
        if(!fs->partition->device_write(dir_entry_offset, buffer, sizeof(buffer)))
            return 0;

        /* check if we deleted the whole entry */
        if(buffer[11] != 0x0f)
            break;

        dir_entry_offset += 32;
    }
#else
    /* mark the directory entry as deleted */
    uint8_t first_char = FAT_DIRENTRY_DELETED;
    if(!fs->partition->device_write(dir_entry_offset, &first_char, 1))
        return 0;
#endif

    /* We deleted the directory entry. The next thing to do is
     * marking all occupied clusters as free.
     */
    return (dir_entry->cluster == 0 || fat_free_clusters(fs, dir_entry->cluster));
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_move_file(struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry, struct fat_dir_struct* parent_new, const char* file_new)
{
    if(!fs || !dir_entry || !parent_new || (file_new && !file_new[0]))
        return 0;
    if(fs != parent_new->fs)
        return 0;

    /* use existing file name if none has been specified */
    if(!file_new)
        file_new = dir_entry->long_name;

    /* create file with new file name */
    struct fat_dir_entry_struct dir_entry_new;
    if(!fat_create_file(parent_new, file_new, &dir_entry_new))
        return 0;

    /* copy members of directory entry which do not change with rename */
    dir_entry_new.attributes = dir_entry->attributes;
    dir_entry_new.cluster = dir_entry->cluster;
    dir_entry_new.file_size = dir_entry->file_size;

    /* make the new file name point to the old file's content */
    if(!fat_write_dir_entry(fs, &dir_entry_new))
    {
        fat_delete_file(fs, &dir_entry_new);
        return 0;
    }
    
    /* delete the old file, but not its clusters, which have already been remapped above */
    dir_entry->cluster = 0;
    if(!fat_delete_file(fs, dir_entry))
        return 0;

    *dir_entry = dir_entry_new;
    return 1;
}
#endif

#if DOXYGEN || FAT_WRITE_SUPPORT
//----------------------------------------------------------------------------
uint8_t fat_create_dir(struct fat_dir_struct* parent, const char* dir, struct fat_dir_entry_struct* dir_entry)
{
    if(!parent || !dir || !dir[0] || !dir_entry)
        return 0;

    /* check if the file or directory already exists */
    while(fat_read_dir(parent, dir_entry))
    {
        if(strcmp(dir, dir_entry->long_name) == 0)
        {
            fat_reset_dir(parent);
            return 0;
        }
    }

    struct fat_fs_struct* fs = parent->fs;

    /* allocate cluster which will hold directory entries */
    cluster_t dir_cluster = fat_append_clusters(fs, 0, 1);
    if(!dir_cluster)
        return 0;

    /* clear cluster to prevent bogus directory entries */
    fat_clear_cluster(fs, dir_cluster);
    
    memset(dir_entry, 0, sizeof(*dir_entry));
    dir_entry->attributes = FAT_ATTRIB_DIR;

    /* create "." directory self reference */
    dir_entry->entry_offset = fs->header.cluster_zero_offset +
                              (offset_t) (dir_cluster - 2) * fs->header.cluster_size;
    dir_entry->long_name[0] = '.';
    dir_entry->cluster = dir_cluster;
    if(!fat_write_dir_entry(fs, dir_entry))
    {
        fat_free_clusters(fs, dir_cluster);
        return 0;
    }

    /* create ".." parent directory reference */
    dir_entry->entry_offset += 32;
    dir_entry->long_name[1] = '.';
    dir_entry->cluster = parent->dir_entry.cluster;
    if(!fat_write_dir_entry(fs, dir_entry))
    {
        fat_free_clusters(fs, dir_cluster);
        return 0;
    }

    /* fill directory entry */
    strncpy(dir_entry->long_name, dir, sizeof(dir_entry->long_name) - 1);
    dir_entry->cluster = dir_cluster;

    /* find place where to store directory entry */
    if(!(dir_entry->entry_offset = fat_find_offset_for_dir_entry(fs, parent, dir_entry)))
    {
        fat_free_clusters(fs, dir_cluster);
        return 0;
    }

    /* write directory to disk */
    if(!fat_write_dir_entry(fs, dir_entry))
    {
        fat_free_clusters(fs, dir_cluster);
        return 0;
    }

    return 1;
}
#endif

//----------------------------------------------------------------------------
#ifdef DOXYGEN
uint8_t fat_delete_dir(struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry);
#endif

//----------------------------------------------------------------------------
#ifdef DOXYGEN
uint8_t fat_move_dir(struct fat_fs_struct* fs, struct fat_dir_entry_struct* dir_entry, struct fat_dir_struct* parent_new, const char* dir_new);
#endif


//----------------------------------------------------------------------------
#if USE_MK3_SD_GET_FS_SIZE==1
offset_t fat_get_fs_size(const struct fat_fs_struct* fs)
{
    if(!fs)
        return 0;

        return (offset_t) (fs->header.fat_size / 2 - 2) * fs->header.cluster_size;
}
#endif

//----------------------------------------------------------------------------
#if USE_MK3_SD_GET_FS_FREE==1
offset_t fat_get_fs_free(const struct fat_fs_struct* fs)
{
    if(!fs)
        return 0;

    uint8_t fat[32];
    struct fat_usage_count_callback_arg count_arg;
    count_arg.cluster_count = 0;
    count_arg.buffer_size = sizeof(fat);

    offset_t fat_offset = fs->header.fat_offset;
    uint32_t fat_size = fs->header.fat_size;
    while(fat_size > 0)
    {
        uintptr_t length = UINTPTR_MAX_UB - 1;
        if(fat_size < length)
            length = fat_size;

        if(!fs->partition->device_read_interval(fat_offset,
                                                fat,
                                                sizeof(fat),
                                                length,

                                                fat_get_fs_free_16_callback,

                                                &count_arg
                                               )
          )
            return 0;

        fat_offset += length;
        fat_size -= length;
    }

    return (offset_t) count_arg.cluster_count * fs->header.cluster_size;
}
#endif

//----------------------------------------------------------------------------
#if USE_MK3_SD_GET_FS_FREE==1
uint8_t fat_get_fs_free_16_callback(uint8_t* buffer, offset_t offset, void* p)
{
    struct fat_usage_count_callback_arg* count_arg = (struct fat_usage_count_callback_arg*) p;
    uintptr_t buffer_size = count_arg->buffer_size;

    for(uintptr_t i = 0; i < buffer_size; i += 2, buffer += 2)
    {
        uint16_t cluster = read16(buffer);
        if(cluster == HTOL16(FAT16_CLUSTER_FREE))
            ++(count_arg->cluster_count);
    }

    return 1;
}
#endif

//----------------------------------------------------------------------------
// FAT.c [ENDE]
//----------------------------------------------------------------------------





//----------------------------------------------------------------------------
// SONSTIGE_UB [START]
//----------------------------------------------------------------------------

typedef enum {          // Rueckgabewerte von READ_FILE
  SDR_READ_OK =0,
  SDR_NO_CARD,
  SDR_PARTITION_ERR,
  SDR_FAT_ERR,
  SDR_DIR_ERR,
  SDR_FILE_NOT_FOUND
}FFR_Status_t;

typedef enum {          // Rueckgabewerte von WRITE_FILE
  SDW_WRITE_OK =0,
  SDW_NO_CARD,
  SDW_PARTITION_ERR,
  SDW_FAT_ERR,
  SDW_DIR_ERR,
  SDW_FILE_WR_ERR,
  SDW_FILE_NOT_FOUND
}FFW_Status_t;

//-------------------------------------------------------------
// Funktionen von UB :
//-------------------------------------------------------------
FFR_Status_t MK3_SD_READ_FILE_IN_ROOT(const char *ptr);
FFW_Status_t MK3_SD_WRITE_FILE_IN_ROOT(const char *ptr);
void MK3_SD_CLOSE_FILE(void);


static struct fat_file_struct* open_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name); 
static uint8_t find_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name, struct fat_dir_entry_struct* dir_entry);


//-------------------------------------------------------------
struct fat_file_struct* open_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name)
{
    struct fat_dir_entry_struct file_entry;
    if(!find_file_in_dir(fs, dd, name, &file_entry))
        return 0;

    return fat_open_file(fs, &file_entry);
} 


//-------------------------------------------------------------
uint8_t find_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name, struct fat_dir_entry_struct* dir_entry)
{
    while(fat_read_dir(dd, dir_entry))
    {
        if(strcmp(dir_entry->long_name, name) == 0)
        {
            fat_reset_dir(dd);
            return 1;
        }
    }

    return 0;
} 


typedef struct {
  struct partition_struct* partition;
  struct fat_fs_struct* fat;
  struct fat_dir_entry_struct root;
  struct fat_dir_struct* dir;
  struct fat_file_struct* file;
}FAST_SD_t;
FAST_SD_t MK3_SD;

//-------------------------------------------------------------
// zum lesen aus einem File im ROOT-Dir
//-------------------------------------------------------------
FFR_Status_t MK3_SD_READ_FILE_IN_ROOT(const char *ptr) {
  FFR_Status_t ret_wert=SDR_READ_OK;

  if(sd_raw_init()!=0)  {
    #if SD_RAW_WRITE_SUPPORT
      MK3_SD.partition = partition_open(sd_raw_read,sd_raw_read_interval,sd_raw_write,sd_raw_write_interval,0);
    #else
      MK3_SD.partition = partition_open(sd_raw_read,sd_raw_read_interval,0,0,0);
    #endif
    if(MK3_SD.partition!=0) {
      MK3_SD.fat = fat_open(MK3_SD.partition);
      if(MK3_SD.fat!=0) {
        fat_get_dir_entry_of_path(MK3_SD.fat, "/", &MK3_SD.root);
        MK3_SD.dir = fat_open_dir(MK3_SD.fat, &MK3_SD.root);
        if(MK3_SD.dir!=0) {
          MK3_SD.file = open_file_in_dir(MK3_SD.fat, MK3_SD.dir, ptr);
          if(MK3_SD.file!=0) {
            ret_wert=SDR_READ_OK;
          }
          else {
            ret_wert=SDR_FILE_NOT_FOUND;
            fat_close_dir(MK3_SD.dir); 
            fat_close(MK3_SD.fat);
            partition_close(MK3_SD.partition);
          }
        }
        else {
          ret_wert=SDR_DIR_ERR;
          fat_close(MK3_SD.fat);
          partition_close(MK3_SD.partition);
        }
      }
      else {
        ret_wert=SDR_FAT_ERR;
        partition_close(MK3_SD.partition);
      }
    }
    else {
      ret_wert=SDR_PARTITION_ERR;
    }
  }
  else {
    ret_wert=SDR_NO_CARD;
  }


  return(ret_wert);
}

//-------------------------------------------------------------
// zum schreiben in ein File im ROOT-Dir
// falls das File noch nicht existiert, wird es angelegt
//-------------------------------------------------------------
FFW_Status_t MK3_SD_WRITE_FILE_IN_ROOT(const char *ptr) {
  FFW_Status_t ret_wert=SDW_WRITE_OK;
  struct fat_dir_entry_struct file_entry;

  #if SD_RAW_WRITE_SUPPORT

  if(sd_raw_init()!=0)  { 
    MK3_SD.partition = partition_open(sd_raw_read,sd_raw_read_interval,sd_raw_write,sd_raw_write_interval,0);
    if(MK3_SD.partition!=0) {
      MK3_SD.fat = fat_open(MK3_SD.partition);
      if(MK3_SD.fat!=0) {
        fat_get_dir_entry_of_path(MK3_SD.fat, "/", &MK3_SD.root);
        MK3_SD.dir = fat_open_dir(MK3_SD.fat, &MK3_SD.root);
        if(MK3_SD.dir!=0) {
          // check ob das File schon vorhanden ist
          if(find_file_in_dir(MK3_SD.fat, MK3_SD.dir, ptr, &file_entry)) {
            // wenn file schon vorhanden
            ret_wert=SDW_WRITE_OK;
          }
          else {
            // wenn file noch nicht vorhanden
            // ein neues anlegen
            if(!fat_create_file(MK3_SD.dir, ptr, &file_entry)) { 
              ret_wert=SDW_FILE_WR_ERR;
              fat_close_dir(MK3_SD.dir); 
              fat_close(MK3_SD.fat);
              partition_close(MK3_SD.partition);
            }
            else {
              sd_raw_sync();
              ret_wert=SDW_WRITE_OK;
            }
          }
          // das File oeffnen
          if(ret_wert==SDW_WRITE_OK) {
            MK3_SD.file = open_file_in_dir(MK3_SD.fat, MK3_SD.dir, ptr);
            if(MK3_SD.file!=0) {
              ret_wert=SDW_WRITE_OK;
            }
            else {
              ret_wert=SDW_FILE_NOT_FOUND;
              fat_close_dir(MK3_SD.dir); 
              fat_close(MK3_SD.fat);
              partition_close(MK3_SD.partition);
            }
          }
        }
        else {
          ret_wert=SDW_DIR_ERR;
          fat_close(MK3_SD.fat);
          partition_close(MK3_SD.partition);
        }
      }
      else {
        ret_wert=SDW_FAT_ERR;
        partition_close(MK3_SD.partition);
      }
    }
    else {
      ret_wert=SDW_PARTITION_ERR;
    }
  }
  else {
    ret_wert=SDW_NO_CARD;
  }
  #endif


  return(ret_wert);
}

//-------------------------------------------------------------
// zum schliessen eines offenen Files
//-------------------------------------------------------------
void MK3_SD_CLOSE_FILE(void) {
  #if SD_RAW_WRITE_SUPPORT
    sd_raw_sync();
  #endif
  fat_close_file(MK3_SD.file);
  fat_close_dir(MK3_SD.dir); 
  fat_close(MK3_SD.fat);
  partition_close(MK3_SD.partition);
}


//----------------------------------------------------------------------------
// SONSTIGE_UB [ENDE]
//----------------------------------------------------------------------------

